#pragma ident  "@(#)tstqcc.c 1.30     98/03/26 SMI"

/*
 *	Copyright (c) 1994, Sun Microsystems, Inc.
 *	All Rights Reserved.
 */

/*
    This program tests the Q.2931 call setup and teardown.
	Syntax:
		# tstqcc interface sap [-c] [-v]
				-c: initiate call
				-v: verbose

    To receive a call:
		# tstqcc sa0 0x100 -v

    To initiate a call:
		# tstqcc sa0 0x100 -c -v
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/stropts.h>
#include <fcntl.h>
#include <sys/errno.h>
#include <sys/time.h>
#include <atm/qcc.h>

#define	Q93B_NAME "q93b"

extern	errno;

int		verbose;

#define	BUFSIZE		0x10000

char	ctlbuf[BUFSIZE];
char	databuf[BUFSIZE];

atm_address_t	src_addr = {
	0x45, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x0f, 0x00, 0x00, 0x00, 0x00,
	0x08, 0x00, 0x20, 0x1a, 0xe1, 0x53, 0x00
};

atm_address_t	dst_addr = {
	0x45, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x0f, 0x00, 0x00, 0x00, 0x00,
	0x08, 0x00, 0x20, 0x1a, 0xb6, 0xb9, 0x00
};

usage(str)
{
	printf("usage: tstqcc interface sap [-c] [-v]\n", str);
	printf("       -c:		initiate call\n");
	printf("       -v:		verbose\n");
	exit(-1);
}

power(x, y)
u_long		x, y;
{
	int		result = 1;

	for (; y; y--) {
		result *= x;
	}
	return (result);
}

u_long
atoh(str)
char		*str;
{
	u_long		result = 0;
	int			i = 0;
	char		*p;

	for (p = str+strlen(str)-1; p >= str; p--, i++) {
		if (*p >= '0' && *p <= '9')
			result = result + ((*p - '0') * power(16, i));
		if (*p >= 'a' && *p <= 'f')
			result = result + ((*p - 'a' + 10) * power(16, i));
		if (*p >= 'A' && *p <= 'F')
			result = result + ((*p - 'A' + 10) * power(16, i));
	}
	return (result);
}

u_long
getdata(str)
char	*str;
{
	u_long	result;

	if (strncmp(str, "0x", 2) == 0)
		result = atoh(&str[2]);
	else
		result = atoi(str);
	return (result);
}

void
pr_time(pt1, pt2)
register hrtime_t		pt1, pt2;
{
	register long	sec;
	hrtime_t		nsec;

	nsec = pt2 - pt1;
	sec = nsec / 1000000000;
	nsec %= 1000000000;

	printf("elapse time = %d.%6lld sec\n", sec, nsec/1000);
}

pr_msgtype(type)
int		type;
{
	switch (type) {
	case QCC_CALL_PROCEEDING:
		printf("CALL_PROCEEDING\n");
		break;
	case QCC_CONNECT:
		printf("CONNECT\n");
		break;
	case QCC_CONNECT_ACK:
		printf("CONNECT_ACK\n");
		break;
	case QCC_SETUP:
		printf("SETUP\n");
		break;
	case QCC_RELEASE:
		printf("RELEASE\n");
		break;
	case QCC_RELEASE_COMPLETE:
		printf("RELEASE_COMPLETE\n");
		break;
	case QCC_RESTART:
		printf("RESTART\n");
		break;
	case QCC_RESTART_ACK:
		printf("RESTART_ACK\n");
		break;
	case QCC_STATUS:
		printf("STATUS\n");
		break;
	case QCC_STATUS_ENQUIRY:
		printf("STATUS_ENQUIRY\n");
		break;
	case QCC_ADD_PARTY:
		printf("ADD_PARTY\n");
		break;
	case QCC_ADD_PARTY_ACK:
		printf("ADD_PARTY_ACK\n");
		break;
	case QCC_ADD_PARTY_REJECT:
		printf("ADD_PARTY_REJECT\n");
		break;
		break;
	case QCC_DROP_PARTY:
		printf("DROP_PARTY\n");
		break;
	case QCC_DROP_PARTY_ACK:
		printf("DROP_PARTY_ACK\n");
		break;
	case QCC_SETUP_ACK:
		printf("SETUP_ACK\n");
		break;
	}
}

int
transmit(fd, ctlbufp, ctllen, databufp, datalen)
int		fd;
char	*ctlbufp;
int		ctllen;
char	*databufp;
int		datalen;
{
	struct strbuf	ctl;
	struct strbuf	data;

	ctl.buf = (char *) ctlbufp;
	ctl.len = ctllen;
	data.buf = (char *) databufp;
	data.len = datalen;
	if (putmsg(fd, &ctl, &data, 0) < 0) {
		printf("putmsg failed, errno=%d\n", errno);
		perror("");
		return (-1);
	}

	if (verbose) {
		qcc_hdr_t	*q93bhdrp = (qcc_hdr_t *) ctlbufp;

		printf("\ntransmit: ");
		pr_msgtype(q93bhdrp->type);
		if (ctllen > 0)
			xdbuf(ctlbufp, ctllen);
		if (datalen > 0)
			xdbuf(databufp, datalen);
	}
	return (0);
}

int
receive(fd, ctlbufp, ctllenp, databufp, datalenp, maxlen)
int			fd;
u_char		*ctlbufp;
int			*ctllenp;
u_char		*databufp;
int			*datalenp;
int			maxlen;
{
	struct strbuf	ctl;
	struct strbuf	data;
	int				flags;

	ctl.buf = (char *) ctlbufp;
	ctl.maxlen = maxlen;
	ctl.len = 0;
	data.buf = (char *) databufp;
	data.maxlen = maxlen;
	data.len = 0;
	flags = 0;
	*ctllenp = 0;
	*datalenp = 0;
	if (getmsg(fd, &ctl, &data, &flags) < 0) {
		printf("getmsg failed, errno=%d\n", errno);
		perror("");
		return (-1);
	}
	*ctllenp = ctl.len;
	*datalenp = data.len;

	if (verbose) {
		qcc_hdr_t	*q93bhdrp = (qcc_hdr_t *) ctlbufp;

		memset(q93bhdrp->ifname + strlen(q93bhdrp->ifname) + 1, 0,
		       QCC_MAX_IFNAME_LEN - strlen(q93bhdrp->ifname) - 1);
		printf("\nreceive: ");
		pr_msgtype(q93bhdrp->type);
		if (*ctllenp > 0)
			xdbuf(ctlbufp, *ctllenp);
		if (*datalenp > 0)
			xdbuf(databufp, *datalenp);
	}
	return (0);
}

outcon(fd, ifname, sap)
int		fd;
char	*ifname;
int		sap;
{
	qcc_hdr_t	q93bhdr;
	int			commandlen;
	int			datalen;
	int			calltag = 0x1234;
	int			callid = 0x5a;
	int			vci = 0x100;
	int			forward_sdusize = 0x2378;
	int			backward_sdusize = 0x2378;
	int			cause;
	struct strbuf	ctl;
	struct strbuf	data;
	hrtime_t	t1, t2, t3;

	callid = 0;

	ctl.maxlen = QCC_MAX_CTL_LEN;
	ctl.buf = ctlbuf;
	data.maxlen = QCC_MAX_DATA_LEN;
	data.buf = databuf;
	t1 = gethrtime();

	qcc_bld_setup(&ctl, &data, ifname, calltag,
					vci, forward_sdusize, backward_sdusize,
					&src_addr, &dst_addr, sap, -1);
	transmit(fd, ctl.buf, ctl.len, data.buf, data.len);

	do {
		receive(fd,
				&q93bhdr,
				&commandlen,
				&databuf[0],
				&datalen,
				BUFSIZE);
	} while (q93bhdr.type != QCC_SETUP_ACK);
	callid = q93bhdr.callid;

	do {
		receive(fd,
				&q93bhdr,
				&commandlen,
				&databuf[0],
				&datalen,
				BUFSIZE);
	} while (q93bhdr.type != QCC_CALL_PROCEEDING);
	data.buf = databuf;
	data.len = datalen;
	qcc_parse_call_proceeding(&data, &vci, NULL);
	printf("parse_call_proceeding: vci=0x%x\n", vci);

	do {
		receive(fd,
				&q93bhdr,
				&commandlen,
				&databuf[0],
				&datalen,
				BUFSIZE);
	} while (q93bhdr.type != QCC_CONNECT);
	data.buf = databuf;
	data.len = datalen;
	qcc_parse_connect(&data, &vci, &forward_sdusize,
						&backward_sdusize, NULL);
	printf("parse_connect: vci=0x%x ", vci);
	printf("forward_sdusize=0x%x backward_sdusize=0x%x\n",
			forward_sdusize, backward_sdusize);

	memset(&q93bhdr, 0, sizeof (qcc_hdr_t));
	strcpy(q93bhdr.ifname, ifname);
	q93bhdr.callid = callid;
	q93bhdr.type = QCC_RELEASE;
	cause = QCC_NORMAL_UNSPECIFIED;
	qcc_bld_release(&ctl, &data, ifname, callid, cause);
	transmit(fd, ctl.buf, ctl.len, data.buf, data.len);

	do {
		receive(fd,
				&q93bhdr,
				&commandlen,
				&databuf[0],
				&datalen,
				BUFSIZE);
	} while (q93bhdr.type != QCC_RELEASE_COMPLETE);

	t2 = gethrtime();
	pr_time(t1, t2);
}

incon(fd, ifname)
int		fd;
char	*ifname;
{
	qcc_hdr_t	q93bhdr;
	int			commandlen;
	int			datalen;
	int			callid;
	int			vci = 0x100;
	int			forward_sdusize = 0x2378;
	int			backward_sdusize = 0x2378;
	int			sap;
	int			cause;
	atm_address_t	src;
	atm_address_t	dst;
	struct strbuf	ctl;
	struct strbuf	data;
	hrtime_t	t1, t2, t3;

	callid = 0;

	for (;;) {
		do {
			receive(fd,
					&q93bhdr,
					&commandlen,
					&databuf[0],
					&datalen,
					BUFSIZE);
		} while (q93bhdr.type != QCC_SETUP);
		data.buf = databuf;
		data.len = datalen;
		qcc_parse_setup(&data, &vci,
						&forward_sdusize,
						&backward_sdusize,
						&src, &dst, &sap, NULL);
		printf("parse_setup: vci=0x%x sap=0x%x ", vci, sap);
		printf("forward_sdusize=0x%x backward_sdusize=0x%x\n",
				forward_sdusize, backward_sdusize);
		callid = q93bhdr.callid;

		t1 = gethrtime();

		ctl.maxlen = QCC_MAX_CTL_LEN;
		ctl.buf = ctlbuf;
		data.maxlen = QCC_MAX_DATA_LEN;
		data.buf = databuf;

		qcc_bld_call_proceeding(&ctl, &data, ifname, callid, vci, -1);
		transmit(fd, ctl.buf, ctl.len, data.buf, data.len);

		memset(&q93bhdr, 0, sizeof (qcc_hdr_t));
		strcpy(q93bhdr.ifname, ifname);
		q93bhdr.callid = callid;
		q93bhdr.type = QCC_CONNECT;

		ctl.maxlen = QCC_MAX_CTL_LEN;
		ctl.buf = ctlbuf;
		data.maxlen = QCC_MAX_DATA_LEN;
		data.buf = databuf;


		qcc_bld_connect(&ctl, &data, ifname, callid,
						vci, forward_sdusize,
						backward_sdusize, -1);
		transmit(fd, ctl.buf, ctl.len, data.buf, data.len);

		do {
			receive(fd,
					&q93bhdr,
					&commandlen,
					&databuf[0],
					&datalen,
					BUFSIZE);
		} while (q93bhdr.type != QCC_CONNECT_ACK);

		do {
			receive(fd,
					&q93bhdr,
					&commandlen,
					&databuf[0],
					&datalen,
					BUFSIZE);
		} while (q93bhdr.type != QCC_RELEASE_COMPLETE);
		data.buf = databuf;
		data.len = datalen;
		qcc_parse_release_complete(&data, &cause);
		printf("parse_release_complete: cause=0x%x\n", cause);

		t2 = gethrtime();
		pr_time(t1, t2);
	}
}

main(argc, argv)
int		argc;
char	**argv;
{
	int		argi;
	int		qfd;
	int		i;
	char	*ifname;
	int		sap;
	int		outgoing = 0;
	char	devname[20];

	if (argc < 3) {
		usage(argv[0]);
		exit(-1);
	}

	argi = 1;
	ifname = argv[argi++];
	sap = getdata(argv[argi++]);

	argc -= 3;
	argv += 3;
	while (argc) {
		if (**argv == '-') {
			switch (argv[0][1]) {
			case 'v':
				verbose = 1;
				break;
			case 'c':
				outgoing = 1;
				break;
			default:
				usage(argv[0]);
				exit(-1);
			}
		} else {
			usage(argv[0]);
			exit(-1);
		}
		argc--;
		argv++;
	}

	memset(devname, 0, sizeof (devname));
	strcpy(devname, "/dev/");
	strcat(devname, Q93B_NAME);
	if ((qfd = open(devname, O_RDWR, 0)) < 0) {
		printf("open %s failed, errno=%d\n", devname, errno);
		perror("open");
		exit(-1);
	}

	if (outgoing)
		outcon(qfd, ifname, sap);
	else {
		if (q_ioc_bind(qfd, sap) < 0) {
			exit(-1);
		}
		incon(qfd, ifname);
	}
}
